/*-------------------------------------------------------------------*
 * This handles all read/write requests to block devices
 *-------------------------------------------------------------------*/
#include <weix/errno.h>
#include <weix/sched.h>
#include <weix/kernel.h>
#include <asm/system.h>
#include <weix/blk.h>
#include <weix/kdebug.h>

#define MAX_BLKDEV 32

#include <weix/extern_func_T_T.h>
/*-------------------------------------------------------------------*
 * global variables
 *-------------------------------------------------------------------*/
/* save all size of block device each major, minor */
int * block_dev_size[MAX_BLKDEV] = { NULL, NULL, };
/*
 * The request-struct contains all necessary data
 * to load a nr of sectors into memory
 * NR_REQUSET == 32
 */
struct request request[NR_REQUEST];
/*
 * used to wait on when there are no free requests
 */
struct my_life_schedule * wait_for_request = NULL;

struct blk_dev_struct block_device[NR_BLK_DEV] = {
	{ NULL, NULL },		/* no_dev */
	{ NULL, NULL },		/* dev mem */
	{ NULL, NULL },		/* dev fd */
	{ NULL, NULL },		/* dev hd */
	{ NULL, NULL },		/* dev ttyx */
	{ NULL, NULL },		/* dev tty */
	{ NULL, NULL }		/* dev lp */
};

/*-------------------------------------------------------------------*
 * lock_buffer
 *-------------------------------------------------------------------*/
static inline void lock_buffer(struct buffer_head * bh)
{
	cli();
	while (bh->b_lock)
		sleep_on(&bh->b_wait);
	bh->b_lock=1;
	sti();
}

/*-------------------------------------------------------------------*
 * unlock_buffer
 *-------------------------------------------------------------------*/
static inline void unlock_buffer(struct buffer_head * bh)
{
	if (!bh->b_lock)
		kprintf("ll_rw_block.c: buffer not locked\n\r");
	bh->b_lock = 0;
	wake_up(&bh->b_wait);
}

/* RO fail safe mechanism */

static long ro_bits[MAX_BLKDEV][8];

int is_read_only(int dev)
{
    int minor,major;

    major = MAJOR(dev);
    minor = MINOR(dev);
    if (major < 0 || major >= MAX_BLKDEV) return 0;
    return ro_bits[major][minor >> 5] & (1 << (minor & 31));
}
/*-------------------------------------------------------------------*
 * add-request adds a request to the linked list.
 * It disables interrupts so that it can muck with the
 * request-lists in peace.
 *
 * Note that swapping requests always go before other requests,
 * and are done in the order they appear.
 *-------------------------------------------------------------------*/
static void add_request(struct blk_dev_struct * dev, struct request * req)
{
	struct request * tmp;

#if 0
	kprintf("add_request: in functions\n");
#endif
	req->next = NULL;
	cli();
	if (req->bh)
		req->bh->b_dirt = 0;

	tmp = dev->current_request;
	if (!tmp) {
		sti();
		dev->current_request = req;
		(dev->request_fn)();
		return;
	}

#if 0
	kprintf("add_request: 2nd\n");
#endif
	for ( ; tmp->next ; tmp=tmp->next) {

		if (!req->bh) {
			if (tmp->next->bh) {
				break;
			} else {
				continue;
			}
		}
		if ((IN_ORDER(tmp,req) ||
		    !IN_ORDER(tmp,tmp->next)) &&
		    IN_ORDER(req,tmp->next)) {
			break;
		}
	}
#if 0
	kprintf("add_request: 3rd\n");
#endif
	req->next=tmp->next;
	tmp->next=req;
	sti();
}

/*-------------------------------------------------------------------*
 * make_request
 *-------------------------------------------------------------------*/
static void make_request(int major,int rw, struct buffer_head * bh)
{
	struct request * req;
	int rw_ahead;

	//kprintf("make_request: in fuctions\n");
/* WRITEA/READA is special case - it is not really needed, so if the */
/* buffer is locked, we just forget about it, else it's a normal read */
	if ((rw_ahead = (rw == READA || rw == WRITEA))) {
		if (bh->b_lock)
			return;
		if (rw == READA)
			rw = READ;
		else
			rw = WRITE;
	}
	if ((rw!=READ) && (rw!=WRITE))
		panic("Bad block dev command, must be R/W/RA/WA");

	lock_buffer(bh);
	if ((rw == WRITE && !bh->b_dirt) || (rw == READ && bh->b_uptodate)) {
		unlock_buffer(bh);
		return;
	}
repeat:
/* we don't allow the write-requests to fill up the queue completely:
 * we want some room for reads: they take precedence. The last third
 * of the requests are only for reads.
 */
	if (rw == READ) {
		req = request+NR_REQUEST;
	} else {
		req = request+((NR_REQUEST*2)/3);
	}
/* find an empty request */
	while (--req >= request) {
		if (req->dev<0) {
			break;
		}
	}
/* if none found, sleep on new requests: check for rw_ahead */
	if (req < request) {
		if (rw_ahead) {
			unlock_buffer(bh);
			return;
		}
		//kprintf("make_request: sleep baby\n");
		sleep_on(&wait_for_request);
		goto repeat;
	}
/* fill up the request-info, and add it to the queue */
	req->dev = bh->b_dev;
	req->cmd = rw;
	req->errors=0;
	req->sector = bh->b_blocknr * (bh->b_size >> 9);
	req->nr_sectors = (bh->b_size >> 9);
	req->buffer = bh->b_data;
	req->waiting = NULL;
	req->bh = bh;
	req->next = NULL;
	add_request(major+block_device,req);
}

/*-------------------------------------------------------------------*
 * ll_rw_page
 *-------------------------------------------------------------------*/
void ll_rw_page(int rw, int dev, int page, char * buffer)
{
	struct request * req;
	unsigned int major = MAJOR(dev);

	if (major >= NR_BLK_DEV || !(block_device[major].request_fn)) {
		kprintf("Trying to read nonexistent block-device\n\r");
		return;
	}
	if (rw!=READ && rw!=WRITE)
		panic("Bad block dev command, must be R/W");
repeat:
	req = request+NR_REQUEST;
	while (--req >= request)
		if (req->dev<0)
			break;
	if (req < request) {
		sleep_on(&wait_for_request);
		goto repeat;
	}
/* fill up the request-info, and add it to the queue */
	req->dev = dev;
	req->cmd = rw;
	req->errors = 0;
	req->sector = page<<3;
	req->nr_sectors = 8;
	req->buffer = buffer;
	req->waiting = current;
	req->bh = NULL;
	req->next = NULL;
	current->state = TASK_UNINTERRUPTIBLE;
	add_request(major+block_device,req);
	schedule();
}	

/*-------------------------------------------------------------------*
 * ll_rw_block
 *-------------------------------------------------------------------*/
void ll_rw_block(int rw, int nr, struct buffer_head * bh[])
{
    unsigned int major;

	nr=0;
    if ((major=MAJOR(bh[0]->b_dev)) >= NR_BLK_DEV || !(block_device[major].request_fn)) {
        kprintf("Trying to read nonexistent block-device\n\r");
        return;
    }
    make_request(major,rw,bh[0]);
}

/*-------------------------------------------------------------------*
 * block_device_init
 *-------------------------------------------------------------------*/
void block_device_init(void)
{
	int i;

	for (i=0; i<NR_REQUEST; i++) {
		request[i].dev = -1;
		request[i].next = NULL;
	}

    /* initiate harddisk */
    hd_init();
}
