/*
 * Hangul console driver for Linux
 *
 * Copyright (C) 1997 Bang Jun-Young <bangjy@geocities.com>
 */

#include <linux/tty.h>
#include <asm/io.h>
#include "selection.h"
#include "console_struct.h"
#include "hanchar.h"

#define GR_PORT_REG 	0x3ce
#define GR_PORT_VAL     0x3cf

extern const unsigned char ascii_font[256][16];
extern unsigned char cho_font[8][20][32];
extern unsigned char joong_font[4][22][32];
extern unsigned char jong_font[4][28][32];
extern const unsigned char ksc5601_johab_table[];

static void eputch(short ch);
static void hputch(hanchar_t h);
static void ksc5601_to_johab(hanchar_t *h);
inline static void mem_overlap(char *dest, const char *src, int size);
unsigned char *graph_mem = 0xa0000;
unsigned short text_color, bg_color;
unsigned short gr_x, gr_y;
int offset, dup_offset;
volatile int dummy;

static const unsigned short offset_y[30] = {
	0x0000, 0x0500, 0x0a00, 0x0f00, 0x1400, 0x1900, 0x1e00, 0x2300,
	0x2800, 0x2d00, 0x3200, 0x3700, 0x3c00, 0x4100, 0x4600, 0x4b00,
	0x5000, 0x5500, 0x5a00, 0x5f00, 0x6400, 0x6900, 0x6e00, 0x7300,
	0x7800, 0x7d00, 0x8200, 0x8700, 0x8c00, 0x9100
};
static unsigned char image_buf[32];

/* Phonetic codes of johab Hangul font */
const char han_font_code[3][30] = {
	/* chosung */
	{
		0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 
		17, 18, 19
	},
	/* joongsung */
	{
		0, 0, 0, 1, 2, 3, 4, 5, 0, 0, 6, 7, 8, 9, 10, 11, 0, 0, 12, 
		13, 14, 15, 16, 17, 0, 0, 18, 19, 20, 21
	},
	/* jongsung */ 
	{
		0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 
		0, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27
	}
};

const char cho_font_type[2][22] = {
	/* Chosung type (bul) used with jongsung */
	{
		0, 5, 5, 5, 5, 5, 5, 5, 5, 6, 7, 7, 7, 6, 6, 7, 7, 7, 6, 6, 
		7, 5
	},
	/* Chosung type (bul) used without jongsung */
	{
		0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 3, 3, 3, 1, 2, 4, 4, 4, 2, 1, 
		3, 0
	}
};
const char joong_font_type[2][20] = {
	/* Joongsung type (bul) used with jongsung */
	{
		0, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 3, 3, 3
	},
	/* Joongsung type (bul) used without jongsung */
	{
		0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1
	},
};
const char jong_font_type[22] = {
	0, 0, 2, 0, 2, 1, 2, 1, 2, 3, 0, 2, 1, 3, 3, 1, 2, 1, 3, 3, 1, 1
};

void scr_writeb(unsigned char val, unsigned char *addr)
{
	*addr = val;
	/* FIXME: not implemented. */
}

void scr_writew(unsigned short val, unsigned short *addr)
{
	static hanchar_t h;
	static unsigned char second_hanchar = 0;
	unsigned short ch;
	unsigned short video_mem_offset;
	static unsigned short last_x, last_y;

	*addr = val;

	ch = val & 0x00ff;
	text_color = (val & 0x0f00) >> 8;
	bg_color = (val & 0x7000) >> 12;
	video_mem_offset = ((unsigned short) addr - 0x8000) / 2;
	gr_x = video_mem_offset % 80;
	gr_y = video_mem_offset / 80;

	/* According to KSC-5601, MSB of each byte must set to 1. */
	if (second_hanchar) {
		if (gr_x - 1 == last_x && last_y == gr_y) {		    
			if (ch & 0x80) {
				h.byte.second = ch;
				gr_x--;
				hputch(h);
			} else {
				/* Second byte is not Hangul. */
				eputch(ch);
			}
		} else {
			/* Cursor has moved, not Hangul. */
			eputch(ch);
		}
		second_hanchar = 0;
	} else if (ch & 0x80) {        
		/* ch is the first byte of Hangul. */
		h.byte.first = ch;
		second_hanchar = 1;
		last_x = gr_x;
		last_y = gr_y;
		eputch(ch);
		return;
	} else { 
		/* ch is a plain character and not Hangul. */
		eputch(ch);
	}
}

void eputch(short ch)
{
	int i;

        /* Clear current position with background color. */
	dup_offset = offset = offset_y[gr_y] + gr_x;
	outw_p(0x0305, GR_PORT_REG);
	outw_p(bg_color << 8, GR_PORT_REG);
	for (i = 0; i < 16; i++) {
		dummy = graph_mem[offset];
		graph_mem[offset] = 0xff;
		offset += 80;
        }
	
        /* Write character at current position with text color. */
        offset = dup_offset;
        outw_p(text_color << 8, GR_PORT_REG);
        for (i = 0; i < 16; i++) {
		dummy = graph_mem[offset];
		graph_mem[offset] = ascii_font[ch][i];
		offset += 80;
        }
        return;
}

void hputch(hanchar_t h)
{    
	unsigned short cho, joong, jong;
	unsigned short cho_type, joong_type, jong_type;
	int i;

	ksc5601_to_johab(&h);
    
	cho = han_font_code[0][h.phon.cho];
	joong = han_font_code[1][h.phon.joong];
	jong = han_font_code[2][h.phon.jong];

	jong_type = jong_font_type[joong];
	if (jong) {
		joong_type = joong_font_type[0][cho];
		cho_type = cho_font_type[0][joong];	
	} else {
		joong_type = joong_font_type[1][cho];
		cho_type = cho_font_type[1][joong];	
	}

	if (cho)
		memcpy(image_buf, cho_font[cho_type][cho], 32);
	else
		memset(image_buf, 0, 32);
	if (joong)
		mem_overlap(image_buf, joong_font[joong_type][joong], 32);
	if (jong)
		mem_overlap(image_buf, jong_font[jong_type][jong], 32);
	
	dup_offset = offset = offset_y[gr_y] + gr_x;
	/* Clear character location with background color */
	outw_p(0x0305, GR_PORT_REG);
	outw_p(bg_color << 8, GR_PORT_REG);
	for (i = 0; i < 16; i++) {
		dummy = graph_mem[offset];
		graph_mem[offset] = 0xff;
		offset++;
		dummy = graph_mem[offset];
		graph_mem[offset] = 0xff;
		offset += 79;
	}

	/* Draw character with text color. */
	offset = dup_offset;
	outw_p(text_color << 8, GR_PORT_REG);
	for (i = 0; i < 32; i += 2) {
		dummy = graph_mem[offset];
		graph_mem[offset] = image_buf[i];
		offset++;
		dummy = graph_mem[offset];
		graph_mem[offset] = image_buf[i+1];
		offset += 79;
	}
}

/* Convert KSC-5601 Hangul code to johab code. */
static void ksc5601_to_johab(hanchar_t *h)
{
    int han_pos;

    han_pos = (h->byte.first - 176) * 94 + h->byte.second - 161;
    h->byte.first = (unsigned short) ksc5601_johab_table[han_pos << 1];
    h->byte.second = (unsigned short) ksc5601_johab_table[(han_pos << 1) + 1];
}

void mem_overlap(char *dest, const char *src, int size)
{
    do
	*dest++ |= *src++;
    while(--size);
}
