/*-------------------------------------------------------------------
 * drivers/char/kbd.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <asm/system.h>
#include <asm/io.h>
#include <weix/signal.h>
#include <weix/sched.h>
#include <weix/interrupt.h>
#include <weix/keyboard.h>
#include <weix/kd.h>
#include <weix/tty.h>

#include <weix/extern_func_T_T.h>

/*-------------------------------------------------------------------
 * defines & global variables
 *-------------------------------------------------------------------*/
#define KEYBOARD_IRQ		1
struct kbd_struct {
    unsigned char ledstate;     /* 3 bits */
    unsigned char default_ledstate;
#define VC_SCROLLOCK    0   /* scroll-lock mode */
#define VC_NUMLOCK  1   /* numeric lock mode */
#define VC_CAPSLOCK 2   /* capslock mode */
    unsigned char lockstate;    /* 4 bits - must be in 0..15 */
#define VC_SHIFTLOCK    KG_SHIFT    /* shift lock mode */
#define VC_ALTGRLOCK    KG_ALTGR    /* altgr lock mode */
#define VC_CTRLLOCK KG_CTRL     /* control lock mode */
#define VC_ALTLOCK  KG_ALT      /* alt lock mode */
    unsigned char modeflags;
#define VC_APPLIC   0   /* application key mode */
#define VC_CKMODE   1   /* cursor key mode */
#define VC_REPEAT   2   /* keyboard repeat */
#define VC_CRLF     3   /* 0 - enter sends CR, 1 - enter sends CRLF */
#define VC_META     4   /* 0 - meta, 1 - meta=prefix with ESC */
#define VC_PAUSE    5   /* pause key pressed - unused */
#define VC_RAW      6   /* raw (scancode) mode */
#define VC_MEDIUMRAW    7   /* medium raw (keycode) mode */
};
#ifndef KBD_DEFMODE
#define KBD_DEFMODE ((1 << VC_REPEAT) | (1 << VC_META))
#endif
#ifndef KBD_DEFLEDS
/*
 * Some laptops take the 789uiojklm,. keys as number pad when NumLock
 * is on. This seems a good reason to start with NumLock off.
 */
#define KBD_DEFLEDS 0
#endif
#ifndef KBD_DEFLOCK
#define KBD_DEFLOCK 0
#endif

/*
 * Translation of escaped scancodes to keysyms.
 * This should be user-settable.
 */
#define E0_BASE 96

#define E0_KPENTER (E0_BASE+0)
#define E0_RCTRL   (E0_BASE+1)
#define E0_KPSLASH (E0_BASE+2)
#define E0_PRSCR   (E0_BASE+3)
#define E0_RALT    (E0_BASE+4)
#define E0_BREAK   (E0_BASE+5)  /* (control-pause) */
#define E0_HOME    (E0_BASE+6)
#define E0_UP      (E0_BASE+7)
#define E0_PGUP    (E0_BASE+8)
#define E0_LEFT    (E0_BASE+9)
#define E0_RIGHT   (E0_BASE+10)
#define E0_END     (E0_BASE+11)
#define E0_DOWN    (E0_BASE+12)
#define E0_PGDN    (E0_BASE+13)
#define E0_INS     (E0_BASE+14)
#define E0_DEL     (E0_BASE+15)
/* BTC */
#define E0_MACRO   (E0_BASE+16)
/* LK450 */
#define E0_F13     (E0_BASE+17)
#define E0_F14     (E0_BASE+18)
#define E0_HELP    (E0_BASE+19)
#define E0_DO      (E0_BASE+20)
#define E0_F17     (E0_BASE+21)
#define E0_KPMINPLUS (E0_BASE+22)
#define E1_PAUSE   (E0_BASE+23)

static unsigned char e0_keys[128] = {
  0, 0, 0, 0, 0, 0, 0, 0,                 /* 0x00-0x07 */
  0, 0, 0, 0, 0, 0, 0, 0,                 /* 0x08-0x0f */
  0, 0, 0, 0, 0, 0, 0, 0,                 /* 0x10-0x17 */
  0, 0, 0, 0, E0_KPENTER, E0_RCTRL, 0, 0,         /* 0x18-0x1f */
  0, 0, 0, 0, 0, 0, 0, 0,                 /* 0x20-0x27 */
  0, 0, 0, 0, 0, 0, 0, 0,                 /* 0x28-0x2f */
  0, 0, 0, 0, 0, E0_KPSLASH, 0, E0_PRSCR,         /* 0x30-0x37 */
  E0_RALT, 0, 0, 0, 0, E0_F13, E0_F14, E0_HELP,       /* 0x38-0x3f */
  E0_DO, E0_F17, 0, 0, 0, 0, E0_BREAK, E0_HOME,       /* 0x40-0x47 */
  E0_UP, E0_PGUP, 0, E0_LEFT, 0, E0_RIGHT, E0_KPMINPLUS, E0_END,/* 0x48-0x4f */
  E0_DOWN, E0_PGDN, E0_INS, E0_DEL, 0, 0, 0, 0,       /* 0x50-0x57 */
  0, 0, 0, 0, 0, 0, 0, 0,                 /* 0x58-0x5f */
  0, 0, 0, 0, 0, 0, 0, 0,                 /* 0x60-0x67 */
  0, 0, 0, 0, 0, 0, 0, E0_MACRO,              /* 0x68-0x6f */
  0, 0, 0, 0, 0, 0, 0, 0,                 /* 0x70-0x77 */
  0, 0, 0, 0, 0, 0, 0, 0                  /* 0x78-0x7f */
};

#define NR_CONSOLES     8
struct kbd_struct kbd_table[NR_CONSOLES];
static struct kbd_struct * kbd = kbd_table;
static volatile uchar kbd_acknowledgment=0;
static volatile uchar kbd_resend=0;
static volatile uchar kbd_overrun=0;
#define W_KBD_SHIFT		1
int kbd_state = 0;
static struct tty_struct * tty = NULL;

/*-------------------------------------------------------------------
 * kb_wait
 *-------------------------------------------------------------------*/
static inline void kb_wait(void)
{
    int i;

    for (i=0; i<0x10000; i++)
        if ((inb_p(0x64) & 0x02) == 0)
            break;
}

/*-------------------------------------------------------------------
 * send_cmd
 *-------------------------------------------------------------------*/
static inline void send_cmd(uchar c)
{
	kb_wait();
	outb(0x64, c);
}

/*-------------------------------------------------------------------
 * put_queue
 *-------------------------------------------------------------------*/
static void put_queue(int ch)
{
    struct tty_queue *qp=NULL;

    if (!tty)
        return;
    qp = &tty->read_q;

	if(!(qp->head < (TTY_BUF_SIZE)))
		qp->head = 0;

    qp->buf[qp->head] = ch;
    INC(qp->head);
}

/*-------------------------------------------------------------------
 * puts_queue
 *-------------------------------------------------------------------*/
static void puts_queue(char *cp)
{
	struct tty_queue *qp=NULL;
	char ch;

    if (!tty)
        return;
    qp = &tty->read_q;

    while ((ch = *(cp++)) != 0) {
        qp->buf[qp->head] = ch;
        INC(qp->head);
    }
}

/*-------------------------------------------------------------------
 * keyboard_interrupt
 *-------------------------------------------------------------------*/
void keyboard_interrupt()
{
    uchar scancode, tmp_scancode, prev_scancode = 0, raw_mode;

	send_cmd(0xAD);		/* Ű带 Ұϰ ϴ. */
	
    scancode = inb(0x60);	/* 8042  scancode нϴ. */

#if 0
	kprintf("|%x", scancode);
#endif
    if (scancode == 0xfa) {			/* ACK ڵ϶ ó */
		kbd_acknowledgment=1;
        goto end_kbd_intr;
    } else if (scancode == 0xfe) {	/* Resend   ó */
		kbd_resend=1;
        goto end_kbd_intr;
    } else if (scancode == 0x0) {	/* Overrun   ó */
		kbd_overrun=1;
        goto end_kbd_intr;
    } else if (scancode == 0xff) {	/* Overrun   ó */
		kbd_overrun=1;
		prev_scancode = 0;
        goto end_kbd_intr;
    }

	tty = &(TTY_TABLE(0));			/* 0   console  tty_struct  ȯ */
	kbd = kbd_table + 0;

	if (scancode == 0x2a || scancode == 0x36) {		/* Shift Ű   */
		prev_scancode = scancode;
		kbd_state = W_KBD_SHIFT;
		goto end_kbd_intr;
	}

	/* ̹  Ű ڵ尡 Ű带     Ȯ
	 *  Ű带    ̸ ȭ鿡 ½Ű ʴ´. */
	tmp_scancode = scancode;
	tmp_scancode &= 0x80;
	if(tmp_scancode) {
		/*  ڵ带 Ѵ. */
		prev_scancode = 0;
		/* Ű   ϴ. */
		kbd_state = 0;
		goto end_kbd_intr;
	}

	/* ĵ ڵ忡 "Release Bit(Ű  ٴ Ʈ)" Ѵ. */
	scancode &= 0x7f;

	/*  PageUp Ű ٸ */
    if(scancode == 0x49) {
        view_prev_screen();
		goto end_kbd_intr;
	}
	/*  PageDown Ű ٸ */
    if(scancode == 0x51) {
        view_next_screen();
		goto end_kbd_intr;
	}

	/* Ų ڵ带 ۿ ֽϴ. */
	put_queue(scancode);

	/* do_tty_interrupt  ȣմϴ. ׷ ȭ鿡 ڰ  ˴ϴ. */

	/*  raw_mode  cooked  ۵Ǿ  ʽϴ. */
	raw_mode = (kbd->modeflags == VC_RAW);
	do_tty_interrupt(0);

end_kbd_intr:
	send_cmd(0xAE);		/* Ű带 밡ϰ մϴ. */
}

/*-------------------------------------------------------------------
 * kbd_bh
 *-------------------------------------------------------------------*/
static void kbd_bh(void * unused)
{
	kprintf("keyboard bottom half function called.. LET ME SEE!!\n");
}

/*-------------------------------------------------------------------
 * kbd_init
 *-------------------------------------------------------------------*/
void kbd_init()
{
    int i;
    struct kbd_struct * kbd;

    kbd = kbd_table + 0;

    for (i = 0 ; i < NR_CONSOLES ; i++,kbd++) {
        kbd->ledstate = KBD_DEFLEDS;
        kbd->default_ledstate = KBD_DEFLEDS;
        kbd->lockstate = KBD_DEFLOCK;
        kbd->modeflags = KBD_DEFMODE;
    }

    bh_base[KEYBOARD_BH].routine = kbd_bh;
    request_irq(KEYBOARD_IRQ,keyboard_interrupt);
    mark_bh(KEYBOARD_BH);
}
