#ifndef _WEIX_FS_H
#define _WEIX_FS_H

#include <weix/types.h>
#include <weix/dirent.h>
#include <weix/ext2_fs_sb.h>
#include <weix/ext2_fs_i.h>
#include <weix/pipe_fs_i.h>
#include <weix/wait.h>
#include <weix/vfs.h>

struct inode {
    dev_t       i_dev;
    unsigned long   i_ino;
    umode_t     i_mode;
    nlink_t     i_nlink;
    uid_t       i_uid;
    gid_t       i_gid;
    dev_t       i_rdev;
    off_t       i_size;
    time_t      i_atime;
    time_t      i_mtime;
    time_t      i_ctime;
    unsigned long   i_blksize;
    unsigned long   i_blocks;
	struct semaphore i_sem;
    struct inode_operations * i_op;
    struct super_block * i_sb;
    struct my_life_schedule * i_wait;
    struct inode * i_next, * i_prev;
    struct inode * i_hash_next, * i_hash_prev;
    struct inode * i_bound_to, * i_bound_by;
    struct inode * i_mount;
    unsigned short i_count;
    unsigned short i_flags;
    unsigned char i_lock;
    unsigned char i_dirt;
    unsigned char i_pipe;
    unsigned char i_seek;
    unsigned char i_update;
    union {
		struct pipe_inode_info pipe_i;
        struct ext2_inode_info ext2_i;
    } u;
};

struct file {
    mode_t f_mode;
    dev_t f_rdev;           /* needed for /dev/tty */
    off_t f_pos;
    unsigned short f_flags;
    unsigned short f_count;
    unsigned short f_reada;
    struct file *f_next, *f_prev;
    struct inode * f_inode;
    struct file_operations * f_op;
};

struct file_operations {
    int (*lseek) 	(struct inode *, struct file *, off_t, int);
    int (*read)		(struct inode *, struct file *, char *, int);
    int (*write) 	(struct inode *, struct file *, char *, int);
    int (*readdir) 	(struct inode *, struct file *, struct dirent *, int);
    int (*select) 	(struct inode *, struct file *, int, select_table *);
    int (*ioctl) 	(struct inode *, struct file *, unsigned int, unsigned long);
    int (*mmap) 	(struct inode *, struct file *, unsigned long, size_t, int, unsigned long);
    int (*open) 	(struct inode *, struct file *);
    void (*release) (struct inode *, struct file *);
    int (*fsync) 	(struct inode *, struct file *);
};


struct inode_operations {
    struct file_operations * default_file_ops;
    int (*create) 	(struct inode *,const char *,int,int,struct inode **);
    int (*lookup) 	(struct inode *,const char *,int,struct inode **);
    int (*link) 	(struct inode *,struct inode *,const char *,int);
    int (*unlink) 	(struct inode *,const char *,int);
    int (*symlink) 	(struct inode *,const char *,int,const char *);
    int (*mkdir) 	(struct inode *,const char *,int,int);
    int (*rmdir) 	(struct inode *,const char *,int);
    int (*mknod) 	(struct inode *,const char *,int,int,int);
    int (*rename) 	(struct inode *,const char *,int,struct inode *,const char *,int);
    int (*readlink) (struct inode *,char *,int);
    int (*follow_link) (struct inode *,struct inode *,int,int,struct inode **);
    int (*bmap) 	(struct inode *,int);
    void (*truncate) (struct inode *);
    int (*permission) (struct inode *, int);
};

/* devices are as follows: (same as minix, so we can use the minix
 * file system. These are major numbers.)
 *
 * 0 - unused (nodev)
 * 1 - /dev/mem
 * 2 - /dev/fd
 * 3 - /dev/hd
 * 4 - /dev/ttyx
 * 5 - /dev/tty
 * 6 - /dev/lp
 * 7 - unnamed pipes
 */

#define IS_BLOCKDEV(x) ((x)==2 || (x)==3)

#define NAME_LEN 14

#define NR_IHASH 131
#define NR_OPEN 20
#define NR_INODE 32
#define NR_FILE 64
#define NR_SUPER 8
#define NR_HASH 307
#define NR_BUFFERS nr_buffers
#define BLOCK_SIZE 1024
#define BLOCK_SIZE_BITS 10
#ifndef NULL
#define NULL ((void *) 0)
#endif 

#define DIR_ENTRIES_PER_BLOCK ((BLOCK_SIZE)/(sizeof (struct dir_entry)))

#define READ 0
#define WRITE 1
#define READA 2     /* read-ahead - don't pause */
#define WRITEA 3    /* "write-ahead" - silly, but somewhat useful */


#define MAJOR(a) (((unsigned)(a))>>8)
#define MINOR(a) ((a)&0xff)

#include <weix/sched.h>

typedef char buffer_block[BLOCK_SIZE];

struct buffer_head {
    char * b_data;          /* pointer to data block (1024 bytes) */
    unsigned short b_dev;       /* device (0 = free) */
	unsigned short b_size;
    unsigned short b_blocknr;   /* block number */
    unsigned char b_uptodate;
    unsigned char b_dirt;       /* 0-clean,1-dirty */
    unsigned char b_count;      /* users using this block */
    unsigned char b_lock;       /* 0 - ok, 1 -locked */
    struct my_life_schedule * b_wait;
    struct buffer_head * b_prev;
    struct buffer_head * b_next;
    struct buffer_head * b_prev_free;
    struct buffer_head * b_next_free;
};

#define MINIX_PIPE_HEAD(inode) (((long *)((inode).i_zone))[0])
#define MINIX_PIPE_TAIL(inode) (((long *)((inode).i_zone))[1])
#define MINIX_PIPE_SIZE(inode) ((MINIX_PIPE_HEAD(inode)-MINIX_PIPE_TAIL(inode))&(PAGE_SIZE-1))
#define MINIX_PIPE_EMPTY(inode) (MINIX_PIPE_HEAD(inode)==MINIX_PIPE_TAIL(inode))
#define MINIX_PIPE_FULL(inode) (MINIX_PIPE_SIZE(inode)==(PAGE_SIZE-1))
#define MINIX_INC_PIPE(head) \
__asm__("incl %0\n\tandl $4095,%0"::"m" (head))

struct super_block {
    dev_t s_dev;
    unsigned long s_blocksize;
    unsigned char s_blocksize_bits;
    unsigned char s_lock;
    unsigned char s_rd_only;
    unsigned char s_dirt;
    struct super_operations *s_op;
    unsigned long s_flags;
    unsigned long s_magic;
    unsigned long s_time;
    struct inode * s_covered;
    struct inode * s_mounted;
    struct wait_queue * s_wait;
	union {
		struct ext2_sb_info ext2_sb;
	} u;
};

struct super_operations {
    void (*read_inode) (struct inode *);
    int (*notify_change) (int flags, struct inode *);
    void (*write_inode) (struct inode *);
    void (*put_inode) (struct inode *);
    void (*put_super) (struct super_block *);
    void (*write_super) (struct super_block *);
    void (*statfs) (struct super_block *, struct statfs *);
    int (*remount_fs) (struct super_block *, int *, char *);
};

struct file_system_type {
    struct super_block *(*read_super) (struct super_block *, void *, int);
    char *name;
    int requires_dev;
};

struct dir_entry {
    unsigned short inode;
    char name[NAME_LEN];
};

/*
 * These are the fs-independent mount-flags: up to 16 flags are supported
 */

#define MS_RDONLY    1 /* mount read-only */
#define MS_NOSUID    2 /* ignore suid and sgid bits */
#define MS_NODEV     4 /* disallow access to device special files */
#define MS_NOEXEC    8 /* disallow program execution */
#define MS_SYNC     16 /* writes are synced at once */
#define MS_REMOUNT  32 /* alter flags of a mounted FS */

/*
 * Flags that can be altered by MS_REMOUNT
 */
#define MS_RMT_MASK (MS_RDONLY)

/*
 * Magic mount flag number. Has to be or-ed to the flag values.
 */
#define MS_MGC_VAL 0xC0ED0000 /* magic flag number to indicate "new" flags */
#define MS_MGC_MSK 0xffff0000 /* magic flag number mask */

/*
 * Note that read-only etc flags are inode-specific: setting some file-system
 * flags just means all the inodes inherit those flags by default. It might be
 * possible to overrride it sevelctively if you really wanted to with some
 * ioctl() that is not currently implemented.
 *
 * Exception: MS_RDONLY is always applied to the entire file system.
 */
#define IS_RDONLY(inode) (((inode)->i_sb) && ((inode)->i_sb->s_flags & MS_RDONLY))
#define IS_NOSUID(inode) ((inode)->i_flags & MS_NOSUID)
#define IS_NODEV(inode) ((inode)->i_flags & MS_NODEV)
#define IS_NOEXEC(inode) ((inode)->i_flags & MS_NOEXEC)
#define IS_SYNC(inode) ((inode)->i_flags & MS_SYNC)


/* the read-only stuff doesn't really belong here, but any other place is
   probably as bad and I don't want to create yet another include file. */

#define BLKROSET 4701 /* set device read-only (0 = read-write) */
#define BLKROGET 4702 /* get read-only status (0 = read_write) */
#define BLKRRPART 4703 /* re-read partition table */
#define BLKGETSIZE 4704 /* return device size */
#define BLKFLSBUF 4705 /* flush buffer cache */

extern struct m_inode inode_table[NR_INODE];
extern struct file file_table[NR_FILE];
extern struct super_block super_blocks[NR_SUPER];
extern struct buffer_head * start_buffer;
extern int nr_buffers;

extern void truncate(struct m_inode * inode);
extern void wait_on(struct m_inode * inode);
extern int create_block(struct m_inode * inode,int block);
extern struct buffer_head * get_hash_table(int dev, int block, int size);
extern struct buffer_head * get_block(int dev, int block, int size);
extern void ll_rw_block(int rw, int, struct buffer_head * bh[]);
extern void brelse(struct buffer_head * buf);
extern struct buffer_head * block_read(int dev,int block, int size);
extern int new_block(int dev);
extern void free_block(int dev, int block);
extern struct m_inode * new_inode(int dev);
extern void free_inode(struct m_inode * inode);

extern inline struct super_block * get_super(int dev)
{
    struct super_block * s;

    for(s = 0+super_blocks;s < NR_SUPER+super_blocks; s++)
        if (s->s_dev == dev)
            return s;
    return NULL;
}

#endif
