/** mm.h
 **
 ** Description:
 ** Memory Management (paging etc.)
 **
 **/

#ifndef __WEIX_MM_H_
#define __WEIX_MM_H_

#include <weix/types.h>
#include <weix/params.h>
#include <weix/sched.h>
#include <weix/kernel.h>
#include <weix/signal.h>
#include <weix/kdebug.h>

#define GFP_BUFFER  0x00
#define GFP_ATOMIC  0x01
#define GFP_USER    0x02
#define GFP_KERNEL  0x03

#define OWNER_NOBODY      -1
#define OWNER_KERNEL       0

#define MAP_RESERVED		(1<<15)
#define MAP_FREE			(0)
#define MAP_USING			(1)


#define read_swap_page(nr,buffer) ll_rw_page(READ,SWAP_DEV,(nr),(buffer));
#define write_swap_page(nr,buffer) ll_rw_page(WRITE,SWAP_DEV,(nr),(buffer));

#define invalidate() \
__asm__ __volatile__("movl %%cr3,%%eax\n\tmovl %%eax,%%cr3": : :"ax")


/*
 * Amount of ram memory (in bytes, 640k-1M not discounted). Currently 8Mb.
 * Don't make this bigger without making sure that there are enough page
 * directory entries (boot/head.s)
 */
#define HIGH_MEMORY 0x800000

/* End of buffer memory. Must be 0xA0000, or > 0x100000, 4096-byte aligned */
#define BUFFER_END  0x200000

/* End of memory. it's start at 0x0. current WEIX support 0M~8M
 * memory management */
#define LOW_MEM     0x0

/* these are not to be changed - thay are calculated from the above */
#define PAGING_MEMORY (HIGH_MEMORY - LOW_MEM)
#define PAGING_PAGES (PAGING_MEMORY/4096)
#define MAP_NR(addr) (((addr)-LOW_MEM)>>12)

#if (PAGING_PAGES < 10)
#error "Won't work"
#endif

typedef struct __region_t
{
  int owner;
  ulong  size;
  struct __region_t *next;
  struct __region_t *prev;
} region_t;

/* page directory entry - 32 bits */
typedef struct __page_dir_entry
{
	byte p : 1;
	byte w : 1;
	byte u : 1;
	byte pwt : 1;
	byte pcd : 1;
	byte a : 1;
	byte tmp_1 : 1;
	byte ps : 1;
	word tmp_2 : 4;
	word pdaddr_12_23 : 12;
	byte pdaddr_24_31;
} __attribute__ ((packed)) page_dir_entry;

typedef struct __page_table_entry 
{
    byte p : 1;
    byte w : 1;
    byte u : 1;
    byte pwt : 1;
    byte pcd : 1;
    byte a : 1;
    byte d : 1;
    byte tmp_1 : 1;
    word tmp_2 : 4;
    word ptaddr_12_23 : 12;
    byte ptaddr_24_31;
} __attribute__ ((packed)) page_table_entry;

struct vm_area_struct {
    struct my_life_schedule * vm_task;       /* VM area parameters */
    unsigned long vm_start;
    unsigned long vm_end;
    unsigned short vm_page_prot;
    struct vm_area_struct * vm_next;    /* linked list */
    struct vm_area_struct * vm_share;   /* linked list */
    struct inode * vm_inode;
    unsigned long vm_offset;
    struct vm_operations_struct * vm_ops;
};

struct vm_operations_struct {
    void (*open)(struct vm_area_struct * area);
    void (*close)(struct vm_area_struct * area);
    void (*nopage)(int error_code,
               struct vm_area_struct * area, unsigned long address);
    void (*wppage)(struct vm_area_struct * area, unsigned long address);
    int (*share)(struct vm_area_struct * from, struct vm_area_struct * to, unsigned long address);
    int (*unmap)(struct vm_area_struct *area, unsigned long, size_t);
};

extern page_dir_entry *PDE;
extern page_table_entry *PTE;

#define REGION_FREE(x) \
  (((((region_t*)(x))->owner) == OWNER_NOBODY) && \
   ((((region_t*)(x))->size) == 0)) 

#define PAGE_ROUND_UP(x) ( (x&0xfff) ? ((x&(~0xfff))+0x1000) : x )
#define PAGE_ROUND(x) ((x) & 0xFFFFF000)

/* linear address to page number */
#define linear2pde(x)   ((x) >> PDE_SHIFT)
#define linear2pte(x)   ((x) >> PTE_SHIFT)

/* page number to address */
#define pde2linear(x)   ((ulong)(x) << PDE_SHIFT)
#define pte2linear(x)   ((ulong)(x) << PTE_SHIFT)

/* Paging */
void setup_paging();
void seting_PDE(page_dir_entry *pd, byte p, byte w, byte u, byte pwt, byte pcd, byte a, byte ps, dword pdaddr);
void seting_PTE(page_table_entry *pt, byte p, byte w, byte u, byte pwt, byte pcd, byte a, byte d, dword ptaddr);
void enable_paging(void);

/* Physical Memory */
ulong setup_memory(ulong availableMemory);
void showUsedMemory(void);
void show_free_memory(void);

/* linux functions */
unsigned long get_free_page(void);
void free_page(unsigned long addr);
int free_page_tables(unsigned long from,unsigned long size);
int copy_page_tables(unsigned long from,unsigned long to,long size);
unsigned long put_page(unsigned long page,unsigned long address);
void un_wp_page(unsigned long * table_entry);
void do_wp_page(unsigned long error_code,unsigned long address);
void write_verify(unsigned long address);
void do_no_page(unsigned long error_code,unsigned long address);
void calc_mem(void);
ulong mem_init(ulong start_mem, ulong );
extern inline volatile void oom(void);
#endif /* __WEIX_MM_H_ */
