/*-------------------------------------------------------------------*
 * kernel/fs/devices.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <weix/sched.h>
#include <weix/kernel.h>
#include <weix/fs.h>
#include <weix/ext2_fs.h>
#include <weix/types.h>
#include <weix/stat.h>
#include <weix/unistd.h>
#include <weix/string.h>
#include <weix/dirent.h>
#include <asm/bitops.h>

#include <weix/kdebug.h>
#include <weix/extern_func_T_T.h>

/*-------------------------------------------------------------------*
 * defines
 *-------------------------------------------------------------------*/
#define MAX_CHRDEV 32
#define ENODEV          19      /* No such device */
#define MAX_BLKDEV 32
#define EBUSY           16      /* Device or resource busy */
#define EINVAL          22      /* Invalid argument */

/*-------------------------------------------------------------------*
 * global variables
 *-------------------------------------------------------------------*/
struct device_struct {
    const char * name;
    struct file_operations * fops;
};

static struct device_struct chrdevs[MAX_CHRDEV] = {
    { NULL, NULL },
};

static struct device_struct blkdevs[MAX_BLKDEV] = {
    { NULL, NULL },
};

/*-------------------------------------------------------------------*
 * get_blkfops
 *-------------------------------------------------------------------*/
struct file_operations * get_blkfops(unsigned int major)
{
    if (major >= MAX_BLKDEV)
        return NULL;
    return blkdevs[major].fops;
}

/*-------------------------------------------------------------------*
 * get_chrfops
 *-------------------------------------------------------------------*/
struct file_operations * get_chrfops(unsigned int major)
{
    if (major >= MAX_CHRDEV)
        return NULL;
    return chrdevs[major].fops;
}

/*-------------------------------------------------------------------*
 * register_chrdev
 *-------------------------------------------------------------------*/
int register_chrdev(unsigned int major, const char * name, struct file_operations *fops)
{
    if (major >= MAX_CHRDEV)
        return -EINVAL;
    if (chrdevs[major].fops)
        return -EBUSY;
    chrdevs[major].name = name;
    chrdevs[major].fops = fops;
    return 0;
}

/*-------------------------------------------------------------------*
 * register_blkdev
 *-------------------------------------------------------------------*/
int register_blkdev(unsigned int major, const char * name, struct file_operations *fops)
{
    if (major >= MAX_BLKDEV)
        return -EINVAL;
    if (blkdevs[major].fops)
        return -EBUSY;
    blkdevs[major].name = name;
    blkdevs[major].fops = fops;
    return 0;
}

/*-------------------------------------------------------------------*
 * unregister_chrdev
 *-------------------------------------------------------------------*/
int unregister_chrdev(unsigned int major, const char * name)
{
    if (major >= MAX_CHRDEV)
        return -EINVAL;
    if (!chrdevs[major].fops)
        return -EINVAL;
    if (strcmp(chrdevs[major].name, name))
        return -EINVAL;
    chrdevs[major].name = NULL;
    chrdevs[major].fops = NULL;
    return 0;
}

/*-------------------------------------------------------------------*
 * unregister_blkdev
 *-------------------------------------------------------------------*/
int unregister_blkdev(unsigned int major, const char * name)
{
    if (major >= MAX_BLKDEV)
        return -EINVAL;
    if (!blkdevs[major].fops)
        return -EINVAL;
    if (strcmp(blkdevs[major].name, name))
        return -EINVAL;
    blkdevs[major].name = NULL;
    blkdevs[major].fops = NULL;
    return 0;
}

/*-------------------------------------------------------------------*
 * chrdev_open
 *-------------------------------------------------------------------*/
int chrdev_open(struct inode * inode, struct file * filp)
{
	int i;
    kprintf("in chrdev_open\n");
	
    i = MAJOR(inode->i_rdev);
	kprintf("inode->i_rdev=%d, i=%d, chrdevs[%d].fops = %08X", inode->i_rdev, i, i, chrdevs[i].fops);
    if (i >= MAX_CHRDEV || !chrdevs[i].fops) {
		kprintf("Oh~~ shit.. chrdev_open error\n");
        return -ENODEV;
	}
    filp->f_op = chrdevs[i].fops;
    if (filp->f_op->open)
        return filp->f_op->open(inode,filp);
    return 0;
}

struct file_operations def_chr_fops = {
    NULL,       /* lseek */
    NULL,       /* read */
    NULL,       /* write */
    NULL,       /* readdir */
    NULL,       /* select */
    NULL,       /* ioctl */
    NULL,       /* mmap */
    chrdev_open,    /* open */
    NULL,       /* release */
};

struct inode_operations chrdev_inode_operations = {
    &def_chr_fops,      /* default file operations */
    NULL,           /* create */
    NULL,           /* lookup */
    NULL,           /* link */
    NULL,           /* unlink */
    NULL,           /* symlink */
    NULL,           /* mkdir */
    NULL,           /* rmdir */
    NULL,           /* mknod */
    NULL,           /* rename */
    NULL,           /* readlink */
    NULL,           /* follow_link */
    NULL,           /* bmap */
    NULL,           /* truncate */
    NULL            /* permission */
};

/*-------------------------------------------------------------------*
 * blkdev_open
 *-------------------------------------------------------------------*/
int blkdev_open(struct inode * inode, struct file * filp)
{
    kprintf("in blkdev_open\n");
    return 0;
}

struct file_operations def_blk_fops = {
    NULL,       /* lseek */
    NULL,       /* read */
    NULL,       /* write */
    NULL,       /* readdir */
    NULL,       /* select */
    NULL,       /* ioctl */
    NULL,       /* mmap */
    blkdev_open,    /* open */
    NULL,       /* release */
};

struct inode_operations blkdev_inode_operations = {
    &def_blk_fops,      /* default file operations */
    NULL,           /* create */
    NULL,           /* lookup */
    NULL,           /* link */
    NULL,           /* unlink */
    NULL,           /* symlink */
    NULL,           /* mkdir */
    NULL,           /* rmdir */
    NULL,           /* mknod */
    NULL,           /* rename */
    NULL,           /* readlink */
    NULL,           /* follow_link */
    NULL,           /* bmap */
    NULL,           /* truncate */
    NULL            /* permission */
};
