/*-------------------------------------------------------------------*
 * super/fs/super.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * super.c contains code to handle the super-block tables.
 */
#include <weix/config.h>
#include <weix/sched.h>
#include <weix/kernel.h>
#include <weix/fs.h>
#include <weix/ext2_fs.h>
#include <weix/types.h>
#include <weix/stat.h>
#include <weix/string.h>
#include <weix/dirent.h>
#include <asm/bitops.h>

#include <weix/kdebug.h>
#include <weix/extern_func_T_T.h> 

/*-------------------------------------------------------------------*
 * preferences
 *-------------------------------------------------------------------*/
struct m_inode * minix_iget(int dev,int nr);
struct buffer_head * block_read(int dev,int block, int size);
void brelse(struct buffer_head * buf);
struct inode * iget(struct super_block * sb,int nr);

/*-------------------------------------------------------------------*
 * global variables
 *-------------------------------------------------------------------*/
/* set_bit uses setb, as gas doesn't recognize setc */
#define set_bit(bitnr,addr) ({ \
int __res; \
__asm__("bt %2,%3;setb %%al":"=a" (__res):"a" (0),"r" (bitnr),"m" (*(addr))); \
__res; })

struct super_block super_blocks[NR_SUPER];
static int nibblemap[] = {4, 3, 3, 2, 3, 2, 2, 1, 3, 2, 2, 1, 2, 1, 1, 0};

/*-------------------------------------------------------------------*
 * ext2_count_free
 *-------------------------------------------------------------------*/
unsigned long ext2_count_free (struct buffer_head * map, unsigned int numchars)
{
    unsigned int i;
    unsigned long sum = 0;

    if (!map)
        return (0);
    for (i = 0; i < numchars; i++)
        sum += nibblemap[map->b_data[i] & 0xf] +
            nibblemap[(map->b_data[i] >> 4) & 0xf];
    return (sum);
}

/*-------------------------------------------------------------------*
 * get_group_desc
 *-------------------------------------------------------------------*/
struct ext2_group_desc * get_group_desc (struct super_block * sb,
                        unsigned int block_group,
                        struct buffer_head ** bh)
{
    unsigned long group_desc;
    unsigned long desc;
    struct ext2_group_desc * gdp;

    if (block_group >= sb->u.ext2_sb.s_groups_count) {
        kprintf("get_group_desc: block_group >= groups_count\nblock_group = %d, groups_count = %lu", block_group, sb->u.ext2_sb.s_groups_count);
		panic("");
	}
    group_desc = block_group / EXT2_DESC_PER_BLOCK(sb);
    desc = block_group % EXT2_DESC_PER_BLOCK(sb);
    if (!sb->u.ext2_sb.s_group_desc[group_desc]) {
        kprintf ("get_group_desc: Group descriptor not loaded\nblock_group = %d, group_desc = %lu, desc = %lu",block_group, group_desc, desc);
		panic("");
	}
    gdp = (struct ext2_group_desc *)
          sb->u.ext2_sb.s_group_desc[group_desc]->b_data;
    if (bh)
        *bh = sb->u.ext2_sb.s_group_desc[group_desc];
    return gdp + desc;
}

/*-------------------------------------------------------------------*
 * read_block_bitmap
 *-------------------------------------------------------------------*/
static void read_block_bitmap (struct super_block * sb,
                   unsigned int block_group,
                   unsigned long bitmap_nr)
{
    struct ext2_group_desc * gdp;
    struct buffer_head * bh;
   
    gdp = get_group_desc (sb, block_group, NULL);
    bh = block_read(sb->s_dev, (gdp->bg_block_bitmap), 1024);
    if (!bh) {
        kprintf("read_block_bitmap: annot read block bitmap\nblock_group = %d, block_bitmap = %lu",block_group, gdp->bg_block_bitmap);
		panic("");
	}
    sb->u.ext2_sb.s_block_bitmap_number[bitmap_nr] = block_group;
    sb->u.ext2_sb.s_block_bitmap[bitmap_nr] = bh;
}

/*-------------------------------------------------------------------*
 * load__block_bitmap
 *-------------------------------------------------------------------*/
static int load__block_bitmap (struct super_block * sb,
                   unsigned int block_group)
{
    int i, j;
    unsigned long block_bitmap_number;
    struct buffer_head * block_bitmap;

    if (block_group >= sb->u.ext2_sb.s_groups_count) {
        kprintf ("load_block_bitmap: block_group >= groups_count\nblock_group = %d, groups_count = %lu",
                block_group, sb->u.ext2_sb.s_groups_count);
		panic("");
	}
    if (sb->u.ext2_sb.s_groups_count <= EXT2_MAX_GROUP_LOADED) {
        if (sb->u.ext2_sb.s_block_bitmap[block_group]) {
            if (sb->u.ext2_sb.s_block_bitmap_number[block_group] !=
                block_group) {
                kprintf ("load_block_bitmap: block_group != block_bitmap_number");
				panic("");
			} else
                return block_group;
        } else {
            read_block_bitmap (sb, block_group, block_group);
            return block_group;
        }
    }

    for (i = 0; i < sb->u.ext2_sb.s_loaded_block_bitmaps &&
            sb->u.ext2_sb.s_block_bitmap_number[i] != block_group; i++)
        ;
    if (i < sb->u.ext2_sb.s_loaded_block_bitmaps &&
        sb->u.ext2_sb.s_block_bitmap_number[i] == block_group) {
        block_bitmap_number = sb->u.ext2_sb.s_block_bitmap_number[i];
        block_bitmap = sb->u.ext2_sb.s_block_bitmap[i];
        for (j = i; j > 0; j--) {
            sb->u.ext2_sb.s_block_bitmap_number[j] =
                sb->u.ext2_sb.s_block_bitmap_number[j - 1];
            sb->u.ext2_sb.s_block_bitmap[j] =
                sb->u.ext2_sb.s_block_bitmap[j - 1];
        }
        sb->u.ext2_sb.s_block_bitmap_number[0] = block_bitmap_number;
        sb->u.ext2_sb.s_block_bitmap[0] = block_bitmap;
    } else {
        if (sb->u.ext2_sb.s_loaded_block_bitmaps < EXT2_MAX_GROUP_LOADED)
            sb->u.ext2_sb.s_loaded_block_bitmaps++;
        else
            brelse (sb->u.ext2_sb.s_block_bitmap[EXT2_MAX_GROUP_LOADED - 1]);
        for (j = sb->u.ext2_sb.s_loaded_block_bitmaps - 1; j > 0;  j--) {
            sb->u.ext2_sb.s_block_bitmap_number[j] =
                sb->u.ext2_sb.s_block_bitmap_number[j - 1];
            sb->u.ext2_sb.s_block_bitmap[j] =
                sb->u.ext2_sb.s_block_bitmap[j - 1];
        }
        read_block_bitmap (sb, block_group, 0);
    }
    return 0;
}

/*-------------------------------------------------------------------*
 * load_block_bitmap
 *-------------------------------------------------------------------*/
static inline int load_block_bitmap (struct super_block * sb,
                     unsigned int block_group)
{
    if (sb->u.ext2_sb.s_loaded_block_bitmaps > 0 &&
        sb->u.ext2_sb.s_block_bitmap_number[0] == block_group)
        return 0;

    if (sb->u.ext2_sb.s_groups_count <= EXT2_MAX_GROUP_LOADED &&
        sb->u.ext2_sb.s_block_bitmap_number[block_group] == block_group &&
        sb->u.ext2_sb.s_block_bitmap[block_group])
        return block_group;

    return load__block_bitmap (sb, block_group);
}

/*-------------------------------------------------------------------*
 * ext2_count_free_blocks
 *-------------------------------------------------------------------*/
unsigned long ext2_count_free_blocks (struct super_block * sb)
{
	return sb->u.ext2_sb.s_es->s_free_blocks_count;
}

/*-------------------------------------------------------------------*
 * block_in_use
 *-------------------------------------------------------------------*/
static inline int block_in_use (unsigned long block,
                struct super_block * sb,
                unsigned char * map)
{
    return test_bit ((block - sb->u.ext2_sb.s_es->s_first_data_block) %
             EXT2_BLOCKS_PER_GROUP(sb), map);
}

/*-------------------------------------------------------------------*
 * read_inode_bitmap
 *-------------------------------------------------------------------*/
static void read_inode_bitmap (struct super_block * sb,
                   unsigned long block_group,
                   unsigned int bitmap_nr)
{
    struct ext2_group_desc * gdp;
    struct buffer_head * bh;

    gdp = get_group_desc (sb, block_group, NULL);
    bh = block_read(sb->s_dev, (gdp->bg_inode_bitmap), 1024);
    if (!bh)
        kprintf ("read_inode_bitmap", "Cannot read inode bitmap\n
                block_group = %lu, inode_bitmap = %lu",
                block_group, gdp->bg_inode_bitmap);
    sb->u.ext2_sb.s_inode_bitmap_number[bitmap_nr] = block_group;
    sb->u.ext2_sb.s_inode_bitmap[bitmap_nr] = bh;
}

/*-------------------------------------------------------------------*
 * load_inode_bitmap
 *-------------------------------------------------------------------*/
int load_inode_bitmap (struct super_block * sb,
                  unsigned int block_group)
{
    int i, j;
    unsigned long inode_bitmap_number;
    struct buffer_head * inode_bitmap;

    if (block_group >= sb->u.ext2_sb.s_groups_count) {
        kprintf ("load_inode_bitmap: 
                block_group >= groups_count\n
                block_group = %d, groups_count = %lu",
                 block_group, sb->u.ext2_sb.s_groups_count);
		panic("");
	}
    if (sb->u.ext2_sb.s_loaded_inode_bitmaps > 0 &&
        sb->u.ext2_sb.s_inode_bitmap_number[0] == block_group)
        return 0;
    if (sb->u.ext2_sb.s_groups_count <= EXT2_MAX_GROUP_LOADED) {
        if (sb->u.ext2_sb.s_inode_bitmap[block_group]) {
            if (sb->u.ext2_sb.s_inode_bitmap_number[block_group] != block_group) {
                kprintf ("load_inode_bitmap: 
                        block_group != inode_bitmap_number");
				panic("");
            } else
                return block_group;
        } else {
            read_inode_bitmap (sb, block_group, block_group);
            return block_group;
        }
    }

    for (i = 0; i < sb->u.ext2_sb.s_loaded_inode_bitmaps &&
            sb->u.ext2_sb.s_inode_bitmap_number[i] != block_group;
         i++)
        ;
    if (i < sb->u.ext2_sb.s_loaded_inode_bitmaps &&
        sb->u.ext2_sb.s_inode_bitmap_number[i] == block_group) {
        inode_bitmap_number = sb->u.ext2_sb.s_inode_bitmap_number[i];
        inode_bitmap = sb->u.ext2_sb.s_inode_bitmap[i];
        for (j = i; j > 0; j--) {
            sb->u.ext2_sb.s_inode_bitmap_number[j] =
                sb->u.ext2_sb.s_inode_bitmap_number[j - 1];
            sb->u.ext2_sb.s_inode_bitmap[j] =
                sb->u.ext2_sb.s_inode_bitmap[j - 1];
        }
        sb->u.ext2_sb.s_inode_bitmap_number[0] = inode_bitmap_number;
        sb->u.ext2_sb.s_inode_bitmap[0] = inode_bitmap;
    } else {
        if (sb->u.ext2_sb.s_loaded_inode_bitmaps < EXT2_MAX_GROUP_LOADED)
            sb->u.ext2_sb.s_loaded_inode_bitmaps++;
        else
            brelse (sb->u.ext2_sb.s_inode_bitmap[EXT2_MAX_GROUP_LOADED - 1]);
        for (j = sb->u.ext2_sb.s_loaded_inode_bitmaps - 1; j > 0; j--) {
            sb->u.ext2_sb.s_inode_bitmap_number[j] =
                sb->u.ext2_sb.s_inode_bitmap_number[j - 1];
            sb->u.ext2_sb.s_inode_bitmap[j] =
                sb->u.ext2_sb.s_inode_bitmap[j - 1];
        }
        read_inode_bitmap (sb, block_group, 0);
    }
    return 0;
}

/*-------------------------------------------------------------------*
 * ext2_check_inodes_bitmap
 *-------------------------------------------------------------------*/
void ext2_check_inodes_bitmap (struct super_block * sb)
{
    struct ext2_super_block * es;
    unsigned long desc_count, bitmap_count, x;
    int bitmap_nr;
    struct ext2_group_desc * gdp;
    int i;

    es = sb->u.ext2_sb.s_es;
    desc_count = 0;
    bitmap_count = 0;
    gdp = NULL;
    for (i = 0; i < sb->u.ext2_sb.s_groups_count; i++) {
        gdp = get_group_desc (sb, i, NULL);
        desc_count += gdp->bg_free_inodes_count;
        bitmap_nr = load_inode_bitmap (sb, i);
        x = ext2_count_free (sb->u.ext2_sb.s_inode_bitmap[bitmap_nr],
                     EXT2_INODES_PER_GROUP(sb) / 8);
        if (gdp->bg_free_inodes_count != x) {
            kprintf ("ext2_check_inodes_bitmap: 
                    Wrong free inodes count in group %d,
                    stored = %d, counted = %lu", i,
                    gdp->bg_free_inodes_count, x);
			panic("");
		}
        bitmap_count += x;
    }
    if (es->s_free_inodes_count != bitmap_count) {
        kprintf ("ext2_check_inodes_bitmap: 
                Wrong free inodes count in super block,
                stored = %lu, counted = %lu",
                es->s_free_inodes_count, bitmap_count);
		panic("");
	}
}

/*-------------------------------------------------------------------*
 * ext2_check_blocks_bitmap
 *-------------------------------------------------------------------*/
void ext2_check_blocks_bitmap (struct super_block * sb)
{
    struct buffer_head * bh;
    struct ext2_super_block * es;
    unsigned long desc_count, bitmap_count, x;
    unsigned long desc_blocks;
    int bitmap_nr;
    struct ext2_group_desc * gdp;
    int i, j;

    es = sb->u.ext2_sb.s_es;
    desc_count = 0;
    bitmap_count = 0;
    gdp = NULL;
    desc_blocks = (sb->u.ext2_sb.s_groups_count + EXT2_DESC_PER_BLOCK(sb) - 1) /
              EXT2_DESC_PER_BLOCK(sb);
    for (i = 0; i < sb->u.ext2_sb.s_groups_count; i++) {
        gdp = get_group_desc (sb, i, NULL);
        desc_count += gdp->bg_free_blocks_count;
        bitmap_nr = load_block_bitmap (sb, i);
        bh = sb->u.ext2_sb.s_block_bitmap[bitmap_nr];

        if (!test_bit (0, bh->b_data)) {
            kprintf ("ext2_check_blocks_bitmap: 
                    Superblock in group %d is marked free", i);
			panic("");
		}
        for (j = 0; j < desc_blocks; j++)
            if (!test_bit (j + 1, bh->b_data)) {
                kprintf ("ext2_check_blocks_bitmap: 
                        Descriptor block #%d in group
                        %d is marked free", j, i);
				panic("");
			}
        if (!block_in_use (gdp->bg_block_bitmap, sb, bh->b_data)) {
            kprintf ("ext2_check_blocks_bitmap: 
                    Block bitmap for group %d is marked free",
                    i);
			panic("");
		}

        if (!block_in_use (gdp->bg_inode_bitmap, sb, bh->b_data)) {
            kprintf ("ext2_check_blocks_bitmap
                    Inode bitmap for group %d is marked free",
                    i);
			panic("");
		}
        for (j = 0; j < sb->u.ext2_sb.s_itb_per_group; j++)
            if (!block_in_use (gdp->bg_inode_table + j, sb, bh->b_data)) {
                kprintf ("ext2_check_blocks_bitmap: 
                        Block #%d of the inode table in
                        group %d is marked free", j, i);
				panic("");
			}

        x = ext2_count_free (bh, sb->s_blocksize);
        if (gdp->bg_free_blocks_count != x) {
            kprintf ("ext2_check_blocks_bitmap: 
                    Wrong free blocks count for group %d,
                    stored = %d, counted = %lu", i,
                    gdp->bg_free_blocks_count, x);
			panic("");
		}
        bitmap_count += x;
    }
    if (es->s_free_blocks_count != bitmap_count) {
        kprintf ("ext2_check_blocks_bitmap: 
                Wrong free blocks count in super block, 
                stored = %lu, counted = %lu",
                es->s_free_blocks_count, bitmap_count);
		panic("");
	}
}

/*-------------------------------------------------------------------*
 * read_inode
 *-------------------------------------------------------------------*/
static void read_inode(struct inode * inode)
{
    if (inode->i_sb && inode->i_sb->s_op && inode->i_sb->s_op->read_inode)
        inode->i_sb->s_op->read_inode(inode);
}

struct inode * __iget(struct super_block * sb, int nr, int crossmntp);

/*-------------------------------------------------------------------*
 * iget
 *-------------------------------------------------------------------*/
struct inode * iget(struct super_block * sb,int nr)
{
    return __iget(sb,nr,1);
}

/*-------------------------------------------------------------------*
 * __iget
 *-------------------------------------------------------------------*/
struct inode * __iget(struct super_block * sb, int nr, int crossnmtp)
{
	struct inode * w_inode;

    /* read ROOT INODE */
    if (!(w_inode = (struct inode *)get_free_page()))
        return NULL;

    w_inode->i_sb = sb;
    w_inode->i_dev = sb->s_dev;
    w_inode->i_ino = nr;
    w_inode->i_flags = sb->s_flags;
    read_inode(w_inode);

	return (w_inode);
}

/*-------------------------------------------------------------------*
 * mount_root
 *-------------------------------------------------------------------*/
void mount_root(void)
{
	int i;
	struct super_block * sb;
	struct inode * inode;

	memset(super_blocks, 0, sizeof(super_blocks));

	for(i=0;i<NR_FILE;i++)
		file_table[i].f_count=0;
	for(sb = &super_blocks[0] ; sb < &super_blocks[NR_SUPER] ; sb++)
		sb->s_dev = 0;
#if 0
	kprintf("ROOT_DEV = %x\n", ROOT_DEV);
#endif
	if (!(sb=do_mount(ROOT_DEV)))
		panic("Unable to mount root");
    if (sb) {
        inode = sb->s_mounted;
        inode->i_count += 3 ;   /* NOTE! it is logically used 4 times, not 1 */
        sb->s_covered = inode;
        sb->s_flags = root_mountflags;
        current->pwd = inode;
        current->root = inode;
        kprintf("WEIX: Mounted Ext2 root filesystem.\n");
        return;
    } else
		panic("can't mount root filesystem\n");
}
