/*-------------------------------------------------------------------*
 * Linux 1.0  ִ request_irq  () 
 * weix ȯ濡 ° Ͽϴ.
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-------------------------------------------------------------------*
 * IRQ's are in fact implemented a bit like signal handlers for the kernel.
 * The same sigaction struct is used, and with similar semantics (ie there
 * is a SA_INTERRUPT flag etc). Naturally it's not a 1:1 relation, but there
 * are similarities.
 *
 * sa_handler(int irq_NR) is the default function called.
 * sa_mask is 0 if nothing uses this IRQ
 * sa_flags contains various info: SA_INTERRUPT etc
 * sa_restorer is the unused
 *-------------------------------------------------------------------*/
#include <weix/errno.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <weix/interrupt.h>
#include <weix/idt.h>
#include <weix/gdt.h>
#include <weix/k_stat.h>
#include <weix/ptrace.h>
#include <weix/signal.h>
#include <weix/kdebug.h>

/*-------------------------------------------------------------------*
 * defines
 *-------------------------------------------------------------------*/
#define CR0_NE 32

/*-------------------------------------------------------------------*
 * gloval variables
 *-------------------------------------------------------------------*/
static unsigned char cache_21 = 0xff;
static unsigned char cache_A1 = 0xff;
unsigned long intr_count = 0;
unsigned long bh_active = 0;
unsigned long bh_mask = 0xFFFFFFFF;
struct bh_struct bh_base[32]; 

/*-------------------------------------------------------------------*
 * disable_irq
 *-------------------------------------------------------------------*/
void disable_irq(unsigned int irq_nr)
{
    unsigned long flags;
    unsigned char mask;

    mask = 1 << (irq_nr & 7);
    save_flags(flags);
    if (irq_nr < 8) {
        cli();
        cache_21 |= mask;
        outb(0x21, cache_21);
        restore_flags(flags);
        return;
    }
    cli();
    cache_A1 |= mask;
    outb(0xA1, cache_A1);
    restore_flags(flags);
}

/*-------------------------------------------------------------------*
 * enable_irq
 *-------------------------------------------------------------------*/
void enable_irq(unsigned int irq_nr)
{
    unsigned long flags;
    unsigned char mask;

    mask = ~(1 << (irq_nr & 7));
    save_flags(flags);
    if (irq_nr < 8) {
        cli();
        cache_21 &= mask;
        outb(0x21, cache_21);
        restore_flags(flags);
        return;
    }
    cli();
    cache_A1 &= mask;
    outb(0xA1, cache_A1);
	restore_flags(flags);
}


/*
 * do_bottom_half() runs at normal kernel priority: all interrupts
 * enabled.  do_bottom_half() is atomic with respect to itself: a
 * bottom_half handler need not be re-entrant.
 */

/*-------------------------------------------------------------------*
 * do_bottom_half
 *-------------------------------------------------------------------*/
asmlinkage void do_bottom_half(void)
{
    unsigned long active;
    unsigned long mask, left;
    struct bh_struct *bh;

    bh = bh_base;
    active = bh_active & bh_mask;
    for (mask = 1, left = ~0 ; left & active ; bh++,mask += mask,left += left) {
        if (mask & active) {
            void (*fn)(void *);
            bh_active &= ~mask;
            fn = bh->routine;
            if (!fn)
                goto bad_bh;
            fn(bh->data);
        }
    }
    return;
bad_bh:
    kprintf ("irq.c:bad bottom half entry\n");
}

/*
 * This builds up the IRQ handler stubs using some ugly macros in irq.h
 *
 * These macros create the low-level assembly IRQ routines that do all
 * the operations that are needed to keep the AT interrupt-controller
 * happy. They are also written to be fast - and to disable interrupts
 * as little as humanly possible.
 *
 * NOTE! These macros expand to three different handlers for each line: one
 * complete handler that does all the fancy stuff (including signal handling),
 * and one fast handler that is meant for simple IRQ's that want to be
 * atomic. The specific handler is chosen depending on the SA_INTERRUPT
 * flag when installing a handler. Finally, one "bad interrupt" handler, that
 * is used when no handler is present.
 */

BUILD_IRQ(FIRST,0,0x01)
BUILD_IRQ(FIRST,1,0x02)
BUILD_IRQ(FIRST,2,0x04)
BUILD_IRQ(FIRST,3,0x08)
BUILD_IRQ(FIRST,4,0x10)
BUILD_IRQ(FIRST,5,0x20)
BUILD_IRQ(FIRST,6,0x40)
BUILD_IRQ(FIRST,7,0x80)
BUILD_IRQ(SECOND,8,0x01)
BUILD_IRQ(SECOND,9,0x02)
BUILD_IRQ(SECOND,10,0x04)
BUILD_IRQ(SECOND,11,0x08)
BUILD_IRQ(SECOND,12,0x10)
BUILD_IRQ(SECOND,13,0x20)
BUILD_IRQ(SECOND,14,0x40)
BUILD_IRQ(SECOND,15,0x80)

/*
 * Pointers to the low-level handlers: first the general ones, then the
 * fast ones, then the bad ones.
 */
static void (*interrupt[16])(void) = {
    IRQ0_interrupt, IRQ1_interrupt, IRQ2_interrupt, IRQ3_interrupt,
    IRQ4_interrupt, IRQ5_interrupt, IRQ6_interrupt, IRQ7_interrupt,
    IRQ8_interrupt, IRQ9_interrupt, IRQ10_interrupt, IRQ11_interrupt,
    IRQ12_interrupt, IRQ13_interrupt, IRQ14_interrupt, IRQ15_interrupt
};

static void (*fast_interrupt[16])(void) = {
    fast_IRQ0_interrupt, fast_IRQ1_interrupt,
    fast_IRQ2_interrupt, fast_IRQ3_interrupt,
    fast_IRQ4_interrupt, fast_IRQ5_interrupt,
    fast_IRQ6_interrupt, fast_IRQ7_interrupt,
    fast_IRQ8_interrupt, fast_IRQ9_interrupt,
    fast_IRQ10_interrupt, fast_IRQ11_interrupt,
    fast_IRQ12_interrupt, fast_IRQ13_interrupt,
    fast_IRQ14_interrupt, fast_IRQ15_interrupt
};

static struct sigaction irq_sigaction[16] = {
    { NULL, 0, 0, NULL }, { NULL, 0, 0, NULL },
    { NULL, 0, 0, NULL }, { NULL, 0, 0, NULL },
    { NULL, 0, 0, NULL }, { NULL, 0, 0, NULL },
    { NULL, 0, 0, NULL }, { NULL, 0, 0, NULL },
    { NULL, 0, 0, NULL }, { NULL, 0, 0, NULL },
    { NULL, 0, 0, NULL }, { NULL, 0, 0, NULL },
    { NULL, 0, 0, NULL }, { NULL, 0, 0, NULL },
    { NULL, 0, 0, NULL }, { NULL, 0, 0, NULL }
};

/*
 * do_IRQ handles IRQ's that have been installed without the
 * SA_INTERRUPT flag: it uses the full signal-handling return
 * and runs with other interrupts enabled. All relatively slow
 * IRQ's should use this format: notably the keyboard/timer
 * routines.
 */

/*-------------------------------------------------------------------*
 * do_IRQ
 *-------------------------------------------------------------------*/
asmlinkage void do_IRQ(int irq, struct pt_regs * regs)
{
    struct sigaction * sa = irq + irq_sigaction;

    kstat.interrupts++;
    sa->sa_handler((int) regs);
}


/*
 * do_fast_IRQ handles IRQ's that don't need the fancy interrupt return
 * stuff - the handler is also running with interrupts disabled unless
 * it explicitly enables them later.
 */

/*-------------------------------------------------------------------*
 * do_fast_IRQ
 *-------------------------------------------------------------------*/
asmlinkage void do_fast_IRQ(int irq)
{
    struct sigaction * sa = irq + irq_sigaction;

    kstat.interrupts++;
    sa->sa_handler(irq);
}


/*-------------------------------------------------------------------*
 * irqaction
 *-------------------------------------------------------------------*/
int irqaction(unsigned int irq, struct sigaction * new_sa)
{
	struct sigaction * sa;
    unsigned long flags;

    if (irq > 15)
        return -EINVAL;
    sa = irq + irq_sigaction;
    if (sa->sa_mask)
        return -EBUSY;
    if (!new_sa->sa_handler)
        return -EINVAL;
    save_flags(flags);
    cli();
    *sa = *new_sa;
    sa->sa_mask = 1;
    if (sa->sa_flags & SA_INTERRUPT)
		interrupt_IDT(&IDT[0x20+irq], KERNEL_CS,(ulong ) fast_interrupt[irq], 0);
    else
		interrupt_IDT(&IDT[0x20+irq], KERNEL_CS,(ulong ) interrupt[irq], 0);
    if (irq < 8) {
        cache_21 &= ~(1<<irq);
        outb(0x21, cache_21);
    } else {
        cache_21 &= ~(1<<2);
        cache_A1 &= ~(1<<(irq-8));
        outb(0x21, cache_21);
        outb(0xA1, cache_A1);
    }
    restore_flags(flags);
    return 0;
}

/*-------------------------------------------------------------------*
 * request_irq
 *-------------------------------------------------------------------*/
int request_irq(unsigned int irq, void (*handler)(int))
{
    struct sigaction sa;

    sa.sa_handler = handler;
    sa.sa_flags = 0;
    sa.sa_mask = 0;
    sa.sa_restorer = NULL;
    return irqaction(irq,&sa);
}
