/*-------------------------------------------------------------------*
 * kernel/mm/memory.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <weix/types.h>
#include <asm/system.h> /* for flush_tlb() */
#include <asm/io.h>
#include <weix/string.h>
#include <weix/signal.h>
#include <weix/page.h>
#include <weix/mm.h> 
#include <weix/malloc.h>

#ifdef DEBUG
#include <weix/kdebug.h>
#endif

#include <weix/extern_func_T_T.h>

/*-------------------------------------------------------------------*
 * global variables
 *-------------------------------------------------------------------*/
static ulong total_pages;
static region_t *list_used;
static region_t *list_free;
unsigned short mem_map [ PAGING_PAGES ] = {0,};

/*-------------------------------------------------------------------*
 * preferences
 *-------------------------------------------------------------------*/
int do_exit(long code);

/*-------------------------------------------------------------------*
 * defines
 *-------------------------------------------------------------------*/
#define copy_page(from,to) \
__asm__("cld ; rep ; movsl"::"S" (from),"D" (to),"c" (1024))


/*
 * Get physical address of first (actually last :-) free page, and mark it
 * used. If no free pages left, return 0.
 */
/*-------------------------------------------------------------------*
 * get_free_page
 *-------------------------------------------------------------------*/
ulong get_free_page(void)
{
	int i;

	/* 8M  2M մϴ.. (PAGING_PAGES-1)    = 8M
	 *						 		   (PAGING_PAGES-1536) = 2M */
	for(i=(PAGING_PAGES-1) ; i > (PAGING_PAGES-1536) ; i--) {
		if(mem_map[i] == MAP_FREE) {
			mem_map[i] = MAP_USING;
			return i*4096;
		} 
	}

	/*  free page ٸ 0 մϴ.
	 * ع swap  Լ ߰ؾ ڽϴ. */
	return 0;
}

/*
 * Free a page of memory at physical address 'addr'. Used by
 * 'free_page_tables()'
 */
/*-------------------------------------------------------------------*
 * free_page
 *-------------------------------------------------------------------*/
void free_page(unsigned long addr)
{
	//kprintf("go into free_page functions \n");
	//kprintf("addr=%08X\n", addr);
    if (addr<LOW_MEM) return;
    if (addr>HIGH_MEMORY)
        panic("trying to free nonexistent page");
    addr -= LOW_MEM;
    addr >>= 12;
    if (mem_map[addr]--) return;
    mem_map[addr]=MAP_FREE;
    panic("trying to free free page");
}

/*
 * This function frees a continuos block of page tables, as needed
 * by 'exit()'. As does copy_page_tables(), this handles only 4Mb blocks.
 */
/*-------------------------------------------------------------------*
 * free_page_tables
 *-------------------------------------------------------------------*/
int free_page_tables(unsigned long from,unsigned long size)
{
    unsigned long *pg_table;
    unsigned long * dir, nr;

#ifdef DEBUG2
	kprintf("go into free_page_tables function\n");
	kprintf("from=%08X, size=%08X\n", from ,size);
#endif
    if (from & 0x3fffff)
        panic("free_page_tables called with wrong alignment");
    if (!from)
        panic("Trying to free up swapper memory space");
    size = (size + 0x3fffff) >> 22;		/* if size = 0xA0000, result is 1. */
    dir = (unsigned long *) (((from>>20) & 0xffc) + PDE_POSITION);
    for ( ; size-->0 ; dir++) {
        if (!(1 & *dir))
            continue;
        pg_table = (unsigned long *) (0xfffff000 & *dir);
        for (nr=0 ; nr<1024 ; nr++) {
            if (1 & *pg_table)
                free_page(0xfffff000 & *pg_table);
            *pg_table = 0;
            pg_table++;
        }
        free_page(0xfffff000 & *dir);
        *dir = 0;
    }
    invalidate();
#ifdef DEBUG2
	kprintf("exit free_page_tables after invalidate()\n");
#endif
    return 0;
}

/*
 *  Well, here is one of the most complicated functions in mm. It
 * copies a range of linerar addresses by copying only the pages.
 * Let's hope this is bug-free, 'cause this one I don't want to debug :-)
 *
 * Note! We don't copy just any chunks of memory - addresses have to
 * be divisible by 4Mb (one page-directory entry), as this makes the
 * function easier. It's used only by fork anyway.
 *
 * NOTE 2!! When from==0 we are copying kernel space for the first
 * fork(). Then we DONT want to copy a full page-directory entry, as
 * that would lead to some serious memory waste - we just copy the
 * first 160 pages - 640kB. Even that is more than we need, but it
 * doesn't take any more memory - we don't copy-on-write in the low
 * 1 Mb-range, so the pages can be shared with the kernel. Thus the
 * special case for nr=xxxx.
 */
/*-------------------------------------------------------------------*
 * copt_page_tables
 *-------------------------------------------------------------------*/
int copy_page_tables(unsigned long from, unsigned long to, long size)
{
    unsigned long *from_page_table;
    unsigned long *to_page_table;
    unsigned long this_page;
    unsigned long *from_dir, *to_dir;
    unsigned long nr, tmptmp;

#if 0
	kprintf("in copy_page_tales\n");
	kprintf("copy page from=%08X, to=%08X, size=%08X\n", from, to, size);
#endif

    if ((from & 0x3fffff) || (to & 0x3fffff)) {
        panic("copy_page_tables called with wrong alignment");
	}

	/* from_dir, to_dir point Page Directory Entry(PDE)
	 * in weix, PDE position in 0x81000 so add PDE_POSITION each
	 * from_dir, to_dir */
    from_dir = (unsigned long *) (((from>>20) & 0xffc) + PDE_POSITION); 
    to_dir = (unsigned long *) (((to>>20) & 0xffc) + PDE_POSITION);
    size = ((unsigned) (size+0x3fffff)) >> 22;

#if 0
	kprintf("from_dir=%08X, to_dir=%08X, size_dir=%08X\n", from_dir, to_dir, size); 
#endif
    for( ; size-->0 ; from_dir++,to_dir++) {
#if 0
		kprintf("in for loop\n");
		kprintf("from_dir=%08X, to_dir=%08X, size=%08X\n", from_dir, to_dir, size);
		kprintf("*from_dir=%08X, *to_dir=%08X\n", *from_dir, *to_dir);
#endif
		/* identify with to_dir and page present bit */
        if (1 & (*to_dir))
            panic("copy_page_tables: already exist");
#if 0
		kprintf("pass 1  to_dir\n");
#endif
		/* identify with to_dir and page present bit */
        if (!(1 & (*from_dir))) {
            continue;
		}
#if 0
		kprintf("pass 1 from_dir\n");
#endif
        from_page_table = (unsigned long *) (0xfffff000 & *from_dir);

#if 0
		kprintf("pass from_page_table\n");
#endif
        if (!(to_page_table = (unsigned long *)get_free_page())) {
            return -1;  
		}

#if 0
		kprintf("pass to_page_table\n");
#endif
		tmptmp = get_free_page();

#if 0
		kprintf("pass tmptmp\n");
#endif
		to_page_table = (ulong *)tmptmp;

        *to_dir = ((unsigned long) to_page_table) | 7;

        nr = ((from==0)? 0xA0 : 1024);

#if 0
		kprintf("second loop ---\n");
		kprintf("from_page_table=%08X,to_page_table=%08X,nr=%d\n", from_page_table, to_page_table, nr);
#endif
        for ( ; nr-- > 0 ; from_page_table++,to_page_table++) {
            this_page = *from_page_table;
            if (!(1 & this_page))
                continue;
            this_page &= ~2;
            *to_page_table = this_page;
            if (this_page > LOW_MEM) {
                *from_page_table = this_page;
                this_page -= LOW_MEM;
                this_page >>= 12;
                mem_map[this_page]++;
            }
        }
    }
#if 0
	kprintf("exited in for loop\n");
#endif
    invalidate();
#if 0
	kprintf("exited in copy_page_tables\n");
#endif
    return 0;
}

/*
 * This function puts a page in memory at the wanted address.
 * It returns the physical address of the page gotten, 0 if
 * out of memory (either when trying to access page-table or
 * page.)
 */
/*-------------------------------------------------------------------*
 * put_page
 *-------------------------------------------------------------------*/
unsigned long put_page(unsigned long page,unsigned long address)
{
    unsigned long tmp, *page_table;

    if (page < LOW_MEM || page > HIGH_MEMORY)
        kprintf("Trying to put page %p at %p\n",page,address);
    if (mem_map[(page-LOW_MEM)>>12] != 1)
        kprintf("mem_map disagrees with %p at %p\n",page,address);
    page_table = (unsigned long *) ((address>>20) & 0xffc);
    if ((*page_table)&1)
        page_table = (unsigned long *) (0xfffff000 & *page_table);
    else {
        if (!(tmp=get_free_page()))
            return 0;
        *page_table = tmp|7;
        page_table = (unsigned long *) tmp;
    }
    page_table[(address>>12) & 0x3ff] = page | 7;
    return page;
}

/*-------------------------------------------------------------------*
 * write_verify
 *-------------------------------------------------------------------*/
void write_verify(unsigned long address)
{
    return;
}

/* return total number of pages */
/*-------------------------------------------------------------------*
 * setup_memory
 *-------------------------------------------------------------------*/
ulong setup_memory(ulong avail_mem)
{
   ulong total_memory;

   /* limit = 512M  :o( */
   if (avail_mem>MAX_MEMORY) total_memory = MAX_MEMORY;
   else total_memory = avail_mem;

   /* number of pages with PAGE_SIZE (=4096) bytes */
   total_pages = (total_memory>>PAGE_SHIFT);

   /* Init data structures */
   list_used = NULL;
   list_free = (region_t *)PHYS_BASE;
   list_free->next = NULL;
   list_free->prev = NULL;
   list_free->owner = OWNER_NOBODY;
   list_free->size = total_memory; 

   return total_pages;
}

/*-------------------------------------------------------------------*
 * mem_init
 *-------------------------------------------------------------------*/
ulong mem_init(ulong start_mem, ulong end_mem)
{
	int i=0;

	/* seting mem_map to RESERVED from 0M ~ 2M */
	for(i=0; i<((start_mem/4096)*2) ; i++)
		mem_map[i] = MAP_RESERVED;

	/* seting mem_map to FREE from 2M ~ 8M */
	for(i=((start_mem/4986)*2); i<PAGING_PAGES; i++)
		mem_map[i] = MAP_FREE;

    /* initiate kmalloc */
    start_mem = kmalloc_init(start_mem, end_mem);

	return (start_mem+sizeof(mem_map));
}

#ifdef DEBUG
/*-------------------------------------------------------------------*
 * __trace_memory
 *-------------------------------------------------------------------*/
static void __trace_memory(region_t *start)
{
   region_t *link = start;
   kprintf("%16s %16s %16s %16s\n","owner","start","end","size");
   while(NULL != link) {
      switch(link->owner) {
      case OWNER_KERNEL: 
			kprintf("%16s ","KERNEL"); 
			break;
      default: 
			kprintf("%16s ","NOBODY");
      }
      kprintf("%016x ",(ulong)link+sizeof(region_t)); 
      kprintf("%016x ",(ulong)link+sizeof(region_t)+link->size); 
      kprintf("%16ld\n",(ulong)(link->size/1024)); 
      link = link->next;
   }
}

void show_used_memory(void) { __trace_memory(list_used); }
void show_free_memory(void) { __trace_memory(list_free); }

#endif

/*-------------------------------------------------------------------*
 * oom
 *-------------------------------------------------------------------*/
inline volatile void oom(void)
{
	return;
}
