#include "kthread.h"
#include "list.h"
#include <asm/io.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <weix/types.h>
#include <weix/string.h>
#include <weix/mm.h>
#include <weix/kdebug.h>

#define THREAD_QUEUE_INITIALIZER { 0, 0 }
#define EFLAGS_IF	(1 << 9)
#define KASSERT( cond )                 \
do {                            \
    if ( !(cond) ) {                    \
    kprintf( "Failed assertion: %s at %s, line %d (thread=%x)",\
        #cond, __FILE__, __LINE__, current_th );   \
    while (1)                   \
       ;                        \
    }                           \
} while (0)
#ifdef NEED_UNDERSCORE
#  define _S(sym) "_" #sym 
#else 
#  define _S(sym) #sym
#endif
#define ASSERT_HELD(mutex)              \
do {                            \
    KASSERT( (mutex)->state == MUTEX_LOCKED );      \
    KASSERT( (mutex)->owner == current_th );   \
} while ( 0 )


enum { MUTEX_UNLOCKED, MUTEX_LOCKED };
struct kernel_thread * current_th;
typedef void (*thread_start_func)( unsigned long arg );

bool interrupt_enabled(void);
void enqueue_thread(struct thread_queue * queue, struct kernel_thread * kthread);
void make_runnable(struct kernel_thread * kthread);
void wake_up_one(struct thread_queue * wait_queue);

static struct thread_queue run_queue = THREAD_QUEUE_INITIALIZER;
static struct thread_queue graveyard_queue = THREAD_QUEUE_INITIALIZER;
static struct thread_queue reaper_wait_queue = THREAD_QUEUE_INITIALIZER;

bool is_thread_queue_empty(struct thread_queue * queue)
{
    return queue->head == 0;
}

void clear_thread_queue(struct thread_queue * queue)
{
	queue->head = queue->tail = 0;
}

void init_cond(struct condition * cond)
{
	clear_thread_queue(&cond->wait_queue);
}

void init_mutex(struct mutex * mutex)
{
	mutex->state = MUTEX_UNLOCKED;
	mutex->owner = 0;
	clear_thread_queue(&mutex->wait_queue);
}

void init_thread(struct kernel_thread * kthread, void * stack_page, int priority, bool detached)
{
	struct kernel_thread * owner = detached ? 0 : current_th;

	memset(kthread, 0, sizeof(*kthread));
    kthread->stack_page = stack_page;
    kthread->esp = ((unsigned long) kthread->stack_page) + 4096; /* 4096 = PAGE_SIZE */
    kthread->num_ticks = 0;
    kthread->priority = priority;
    kthread->user_con = 0;
    kthread->owner = owner;

	kthread->ref_count = detached ? 1 : 2;
	kthread->alive = TRUE;

	init_mutex(&kthread->join_lock);
	init_cond(&kthread->join_cond);
}

struct kernel_thread * create_thread(int priority, bool detached)
{
	struct kernel_thread * kthread;
	void * stack_page = 0;

	kthread = (struct kernel_thread *) get_free_page();
	if(kthread)
		stack_page = (void *) get_free_page();

	if(!kthread)
		return 0;
	if(!stack_page) {
		return 0;
	}

	init_thread(kthread, stack_page, priority, detached);
	return kthread;
}

__inline__ void push(struct kernel_thread * kthread, ulong value)
{
	kthread->esp -= 4;
	*((ulong *) kthread->esp) = value;
}

void restore_thread(void)
{
	kprintf("called restore_thread\n");
}

void remove_thread(struct thread_queue * queue, struct kernel_thread * kthread)
{
	REMOVE_NODE(thread_queue, queue, kthread);
}

__inline__ struct kernel_thread * find_best(struct thread_queue * queue)
{
	struct kernel_thread * kthread = queue->head, * best = 0;

	while(kthread != 0) {
		if(best == 0 || kthread->priority > best->priority)
			best = kthread;
		kthread = NODE_NEXT(thread_queue, kthread);
	}

	return best;
}

struct kernel_thread * get_next_runnable(void)
{
	struct kernel_thread * best = find_best(&run_queue);

	KASSERT( best != 0 );

	remove_thread(&run_queue, best);
	return best;
}

void switch_to_thread(struct kernel_thread * runnable)
{
	kprintf("called switch_to_thread\n");
}

void kthread_schedule(void)
{
	struct kernel_thread * runnable;

	KASSERT(!interrupt_enabled());

	runnable = get_next_runnable();

	switch_to_thread(runnable);
}

void kthread_wait(struct thread_queue * wait_queue)
{
	struct kernel_thread * t_current = current_th;

	KASSERT(!interrupt_enabled());

	enqueue_thread(wait_queue, t_current);

	kthread_schedule();
}

__inline__ void mutex_lock_lmp(struct mutex * mutex)
{
	KASSERT(mutex->owner != current_th);

	while (mutex->state == MUTEX_LOCKED) {
		kthread_wait(&mutex->wait_queue);
	}

	mutex->state = MUTEX_UNLOCKED;
	mutex->owner = current_th;
}
void mutex_lock(struct mutex * mutex)
{
	cli();
	
	mutex_lock_lmp(mutex);

	sti();
}

__inline__ void mutex_unlock_lmp(struct mutex * mutex)
{
	KASSERT((mutex)->state == MUTEX_LOCKED);

	KASSERT(mutex->owner != current_th);

	mutex->state = MUTEX_UNLOCKED;
	mutex->owner = 0;

	wake_up_one(&mutex->wait_queue);
}
void mutex_unlock(struct mutex * mutex)
{
    cli();

    mutex_unlock_lmp(mutex);

    sti();
}

void wake_up_one(struct thread_queue * wait_queue)
{
	struct kernel_thread * best;

	KASSERT(!interrupt_enabled());

	best = find_best(wait_queue);

	if(best != 0) {
		remove_thread(wait_queue, best);
		make_runnable(best);
	}
}
void cond_broadcast(struct condition * cond)
{
	cli();
	wake_up_one(&cond->wait_queue);
	sti();
}

void wake_up(struct thread_queue * wait_queue)
{
	struct kernel_thread * kthread = wait_queue->head, * next;

	KASSERT(!interrupt_enabled());

	while(kthread != 0) {
		next = NODE_NEXT(thread_queue, kthread);
		make_runnable(kthread);
		kthread = next;
	}

	clear_thread_queue(wait_queue);
}

void reap_thread(struct kernel_thread * kthread)
{
	enqueue_thread(&graveyard_queue, kthread);
	wake_up(&reaper_wait_queue);
}
void detach_thread(struct kernel_thread * kthread)
{
	KASSERT(!interrupt_enabled());
	KASSERT(kthread->ref_count > 0);

	--kthread->ref_count;
	if(kthread->ref_count == 0)
		reap_thread(kthread);
}
void shutdown_thread(void)
{
	struct kernel_thread * runnable;
	struct kernel_thread * t_current = current_th;

	if(t_current->owner != 0) {
		mutex_lock(&t_current->join_lock);
		t_current->alive = FALSE;
		cond_broadcast(&t_current->join_cond);
		mutex_unlock(&t_current->join_lock);
	}

	cli();

	runnable = get_next_runnable();

	detach_thread(t_current);

	__asm__ __volatile__ (
		"jmp " _S(restore_thread)
	:
	: "a" (runnable)
	);
}

ulong get_current_eflags(void)
{
	ulong eflags;
	__asm__ (
		"pushfl \n\t"
		"popl	%%eax\n\t"
		"movl	%%eax, %0\n\t"
	: "=m" (eflags)
	);
	return eflags;
}

void launch_thread(void)
{
	sti();
}

void setup_kernel_thread_next(struct kernel_thread * kthread, thread_start_func start_func, ulong arg)
{
	ulong eflags;

	push(kthread, arg);
	push(kthread, (ulong) &shutdown_thread);
	push(kthread, (ulong) start_func);
	
	eflags = get_current_eflags();
	eflags &= ~(EFLAGS_IF);
	push(kthread, eflags);
	
	push(kthread, KERNEL_CS);
	push(kthread, (ulong ) &launch_thread);

	push(kthread, 0);
	push(kthread, 0);

	/* this is Push_General_Registers() */
    push(kthread, 0 ); // eax
    push(kthread, 0 ); // ebx
    push(kthread, 0 ); // edx
    push(kthread, 0 ); // edx
    push(kthread, 0 ); // esi
    push(kthread, 0 ); // edi
    push(kthread, 0 ); // ebp

	push(kthread, KERNEL_DS);
	push(kthread, 0);
	push(kthread, 0);
	push(kthread, 0);
}

bool interrupt_enabled(void)
{
	ulong eflags = get_current_eflags();
	return (eflags & EFLAGS_IF) != 0;
}

void enqueue_thread(struct thread_queue * queue, struct kernel_thread * kthread)
{
	ADD_NODE_BACK(thread_queue, queue, kthread);
}

void make_runnable(struct kernel_thread * kthread)
{
	KASSERT(!interrupt_enabled());
	enqueue_thread(&run_queue, kthread);
}

void make_runnable_atomic(struct kernel_thread * kthread)
{
	cli();
	make_runnable(kthread);
	sti();
}

struct kernel_thread * start_kernel_thread(thread_start_func start_func, ulong arg, int priority, bool detached)
{
	struct kernel_thread * kthread = create_thread(priority, detached);

	if(kthread) {
		setup_kernel_thread_next(kthread, start_func, arg);
		make_runnable_atomic(kthread);
	}

	return kthread;
}

void yield(void)
{
	cli();
	make_runnable(current_th);
	kthread_schedule();
	sti();
}

void idle(ulong arg)
{
	while(TRUE)
		yield();
}

void destroy_user_context(struct user_context * user_con)
{
	KASSERT(user_con->ref_count == 0);

	kprintf("let's make free descriptor\n");

	cli();
	kprintf("let's make free memory\n");
	sti();
}

void detach_user_context(struct kernel_thread * kthread)
{
	struct user_context * old = kthread->user_con;

	kthread->user_con = 0;

	if(old != 0) {
		--old->ref_count;
		if(old->ref_count == 0)
			destroy_user_context(old);
	}
}
void destroy_thread(struct kernel_thread * kthread)
{
	detach_user_context(kthread);

	cli();
	kprintf("let's make free_page\n");
	sti();
}

void reaper(ulong arg)
{
	struct kernel_thread * kthread;

	cli();
	while(TRUE) {
		if((kthread = graveyard_queue.head) == 0)
			kthread_wait(&reaper_wait_queue);
		else {
			clear_thread_queue(&graveyard_queue);
			sti();
			yield();
			while(kthread != 0) {
				struct kernel_thread * next = NODE_NEXT(thread_queue, kthread);
				destroy_thread(kthread);
				kthread = next;
			}
			cli();
		}
	}
}
void kthread_init(void)
{
	struct kernel_thread * main_th = (struct kernel_thread *) (1024 * 1024);

	KASSERT(is_thread_queue_empty( &run_queue ) );

	init_thread(main_th, (void *)((1024*1024)+4096), 5, TRUE);
	current_th = main_th;

	start_kernel_thread(idle, 0, 0, TRUE);
	start_kernel_thread(reaper, 0, 5, TRUE);
}

void cond_wait(struct condition * cond, struct mutex * mutex)
{
	ASSERT_HELD(mutex);

	cli();
	mutex_unlock_lmp(mutex);
	kthread_wait(&cond->wait_queue);
	mutex_lock_lmp(mutex);
	sti();
}
void join(struct kernel_thread * kthread)
{
	KASSERT(kthread->owner == current_th);

	mutex_lock(&kthread->join_lock);
	while(kthread->alive) {
		cond_wait(&kthread->join_cond, &kthread->join_lock);
	}
	mutex_unlock(&kthread->join_lock);

	cli();
	detach_thread(kthread);
	sti();
}

void test_join(ulong arg)
{
	kprintf("This is only a test...\n");
}

void kthread_exit(void)
{
	shutdown_thread();
}

void kthread_test(void)
{
	struct kernel_thread * child;

	kprintf("Starting child... \n");
	child = start_kernel_thread(&test_join, 0, 5, FALSE);
	join(child);
	kprintf("Child has exited\n");

	kthread_exit();
}
